"""Logic for compiling requirements"""
import logging
import sys

import pkg_resources
import six

import qer.dists
import qer.metadata
import qer.repos.pypi
import qer.utils
from qer.utils import parse_requirement, merge_requirements
import qer.repos.repository

from qer.dists import DistributionCollection, RequirementsFile
from qer.repos.repository import NoCandidateException

BLACKLIST = [
    'setuptools'
]

MAX_DOWNGRADE = 3

LOG = logging.getLogger('qer.compile')


def compile_roots(node, source, repo, dists, depth=1):
    """

    Args:
        node (qer.dists.DependencyNode):
        source (DependencyNode):
        repo (Repository):
        dists (DistributionCollection):
        depth:
        verbose:

    Returns:

    """
    logger = logging.LoggerAdapter(LOG, dict(depth=depth))
    logger.debug('Processing node %s', node)

    if node.metadata is not None:

        if node.metadata.meta:
            for req in list(node.dependencies):
                compile_roots(req, node, repo, dists, depth=depth + 1)
        else:
            logger.info('Reusing dist %s %s', node.metadata.name, node.metadata.version)
    else:
        spec_req = node.build_constraints()
        first_failure = None
        original_metadata = None

        for attempt in range(MAX_DOWNGRADE):
            try:
                dist, cached = repo.get_candidate(spec_req)
                source_repo = repo.source_of(spec_req)

                logger.debug('Acquired candidate %s [%s] (%s)', dist, source_repo, 'cached' if cached else 'download')

                nodes_to_recurse = set()
                metadata = None
            except NoCandidateException as no_candidate_ex:
                if attempt == 0:
                    raise no_candidate_ex
                else:
                    break

            try:
                if isinstance(dist, qer.metadata.DistInfo):
                    metadata = dist
                else:
                    metadata = qer.metadata.extract_metadata(dist, origin=source_repo)

                # Save off the original metadata for better error information
                if original_metadata is None:
                    original_metadata = metadata

                nodes_to_recurse = dists.add_dist(metadata, source, source.dependencies[node], repo=source_repo)
                if nodes_to_recurse:
                    for recurse_node in nodes_to_recurse:
                        for req in list(recurse_node.dependencies):
                            compile_roots(req, recurse_node, repo, dists, depth=depth + 1)
                first_failure = None
                break
            except qer.metadata.MetadataError as meta_error:
                logger.warning('The metadata could not be processed for %s (%s)', node.key, meta_error)
                ex = sys.exc_info()
                spec_req = merge_requirements(spec_req,
                                              parse_requirement('{}!={}'.format(meta_error.name, meta_error.version)))
            except NoCandidateException as no_candidate_ex:
                logger.debug('Could not use candidate because some of its dependencies could not be satisfied (%s)',
                             no_candidate_ex)
                ex = sys.exc_info()
                spec_req = merge_requirements(spec_req,
                                              parse_requirement('{}!={}'.format(metadata.name, metadata.version)))

            if first_failure is None:
                first_failure = ex

            for node in nodes_to_recurse:
                dists.remove_dists(node, remove_upstream=False)

        if first_failure is not None:
            if original_metadata is not None:
                nodes_to_recurse = dists.add_dist(original_metadata, source, source.dependencies[node])
                if nodes_to_recurse:
                    for recurse_node in nodes_to_recurse:
                        for req in list(recurse_node.dependencies):
                            compile_roots(req, recurse_node, repo, dists, depth=depth + 1)
            six.reraise(*first_failure)


def perform_compile(input_reqs, repo, constraint_reqs=None):
    """
    Perform a compilation using the given inputs and constraints

    Args:
        input_reqs (dict[str, list[pkg_resources.Requirement]]):
            List of mapping of input requirements. If provided a mapping,
            requirements will be kept separate during compilation for better
            insight into the resolved requirements
        repo (qer.repos.Repository): Repository to use as a source of
            Python packages.
        constraint_reqs (dict[str, list[pkg_resources.Requirement]] or None): Constraints to use
            when compiling
    Returns:
        tuple[DistributionCollection, set[DependencyNode], set[DependencyNode]], the solution and the constraints node used for
            the solution
    """
    results = qer.dists.DistributionCollection()

    constraint_nodes = set()
    nodes = set()
    if constraint_reqs is not None:
        for constraint_source in constraint_reqs:
            constraint_node = results.add_dist(RequirementsFile(constraint_source, constraint_reqs[constraint_source]), None, None)
            constraint_nodes |= constraint_node
            nodes |= constraint_node

    for req_source in input_reqs:
        nodes |= results.add_dist(RequirementsFile(req_source, input_reqs[req_source]), None, None)

    try:
        for node in nodes:
            compile_roots(node, None, repo, dists=results)
    except NoCandidateException as ex:
        ex.results = results
        raise

    return results, nodes - constraint_nodes, constraint_nodes



import os
from os.path import expanduser
import sys

from six.moves import configparser


def _get_config_paths():
    # Try user specific config file first
    if sys.platform.startswith('linux2'):
        home = expanduser("~")
        user_config = os.path.join(home, '.config', 'pip', 'pip.conf')
        site_config = '/etc/pip.conf'
    elif sys.platform == 'win32':
        user_config = os.path.join(os.getenv('LOCALAPPDATA'), 'pip', 'pip.ini')
        site_config = os.path.join(os.getenv('PROGRAMDATA'), 'pip', 'pip.ini')
    else:
        return ()

    return site_config, user_config


def read_pip_default_index():
    config_files = _get_config_paths()

    config = configparser.ConfigParser()
    config.read(config_files)

    try:
        return config.get('global', 'index-url')
    except configparser.NoSectionError:
        return None



from __future__ import print_function

import logging
import sys

import six
from six.moves import map as imap

import qer.dists
import qer.utils

from qer.repos.repository import Repository, Candidate, DistributionType, RequiresPython


class SolutionRepository(Repository):
    def __init__(self, filename, allow_prerelease=None):
        super(SolutionRepository, self).__init__(allow_prerelease=allow_prerelease)
        self.filename = filename
        self.solution = load_from_file(self.filename)
        self._logger = logging.getLogger('qer.repository.solution')

    def __repr__(self):
        return '--solution {}'.format(self.filename)

    @property
    def logger(self):
        return self._logger

    def get_candidates(self, req):
        try:
            node = self.solution[req.name]
            candidate = Candidate(
                node.key,
                node.metadata,
                node.metadata.version,
                RequiresPython(None),
                'any',
                None,
                DistributionType.SOURCE)
            candidate.preparsed = node.metadata
            return [candidate]
        except KeyError:
            return []

    def resolve_candidate(self, candidate):
        return candidate.filename, True

    def close(self):
        pass


def load_from_file(filename):
    result = qer.dists.DistributionCollection()

    if filename == '-':
        reqfile = sys.stdin
    else:
        reqfile = open(filename)

    for line in reqfile.readlines():
        req_part, _, source_part = line.partition('#')
        req_part = req_part.strip()
        if not req_part:
            continue

        req = qer.utils.parse_requirement(req_part)
        source_part = source_part.strip()

        if source_part[0] == '[':
            _, _, source_part = source_part.partition('] ')
        sources = source_part.split(', ')

        pkg_names = imap(lambda x: x.split(' ')[0], sources)
        constraints = imap(lambda x: x.split(' ')[1].replace('(', '').replace(')', '') if '(' in x else None, sources)

        version = qer.utils.parse_version(list(req.specifier)[0].version)
        metadata = qer.dists.DistInfo(req.name, version, [])
        result.add_dist(metadata, None, req)

        for name, constraints in zip(pkg_names, constraints):
            if name and not (name.endswith('.txt') or name.endswith('.out')):
                constraint_req = qer.utils.parse_requirement(name)
                result.add_dist(constraint_req.name, None, constraint_req)
                reverse_dep = result[name]
            else:
                reverse_dep = None
            result.add_dist(metadata.name,
                            reverse_dep,
                            qer.utils.parse_requirement('{}{}{}'.format(metadata.name,
                                                                        ('[' + ','.join(req.extras) + ']') if req.extras else '',
                                                                        constraints if constraints else '')))

    if reqfile is not sys.stdin:
        reqfile.close()

    nodes_to_remove = []
    for node in result:
        if node.metadata is not None:
            try:
                requirements = [value for dep_node, value in six.iteritems(node.dependencies)
                                if dep_node.metadata.name != node.metadata.name]
                if node.extra:
                    requirements = [qer.utils.parse_requirement('{} ; extra=="{}"'.format(req, node.extra))
                                    for req in requirements]
                node.metadata.reqs.extend(requirements)
            except Exception:
                print('Error while processing requirement {}'.format(node), file=sys.stderr)
                raise
        else:
            nodes_to_remove.append(node)

    for node in nodes_to_remove:
        try:
            del result.nodes[node.key]
        except KeyError:
            pass
    return result



from __future__ import print_function

import contextlib
import functools
import imp
import io
import logging
import os
import sys
import tarfile
import tempfile
import types
import zipfile
import abc
from contextlib import closing

# This block is imported for the benefit of setup.py's.
import setuptools
import distutils.core
import distutils.extension
import distutils.command.build_ext
import codecs

import pkg_resources
import six
from six.moves import StringIO

from qer import utils
from qer.dists import DistInfo


class MetadataError(Exception):
    def __init__(self, name, version, ex):
        self.name = name
        self.version = version
        self.ex = ex

    def __str__(self):
        return 'Failed to parse metadata for package {} ({}) - {}: {}'.format(self.name, self.version, self.ex.__class__.__name__, str(self.ex))


class Extractor(six.with_metaclass(abc.ABCMeta, object)):
    @abc.abstractmethod
    def names(self):
        pass

    @abc.abstractmethod
    def open(self, filename, mode='r', encoding=None):
        pass

    @abc.abstractmethod
    def close(self):
        pass

    def relative_opener(self, fake_root, directory):
        def inner_opener(filename, *args, **kwargs):
            archive_path = filename
            if os.path.isabs(filename):
                if filename.startswith(fake_root):
                    archive_path = os.path.relpath(filename, fake_root)
                else:
                    return self.open(filename, *args, **kwargs)
            return self.open(directory + '/' + archive_path, *args, **kwargs)
        return inner_opener

    def contents(self, name):
        return self.open(name, encoding='utf-8').read()


def parse_source_filename(full_filename):
    filename = full_filename
    filename = filename.replace('.tar.gz', '')
    filename = filename.replace('.zip', '')
    filename = filename.replace('.tgz', '')

    dash_parts = filename.split('-')
    version_start = None
    for idx, part in enumerate(dash_parts):
        if part[0].isdigit():
            version_start = idx
            break

    if version_start is None:
        return os.path.basename(full_filename), None

    if version_start == 0:
        raise ValueError('Package name missing: {}'.format(full_filename))

    pkg_name = '-'.join(dash_parts[:version_start])
    version = utils.parse_version('-'.join(dash_parts[version_start:]))

    return pkg_name, version


class NonExtractor(Extractor):
    def __init__(self, path):
        self.path = path
        self.io_open = io.open

    def names(self):
        parent_dir = os.path.abspath(os.path.join(self.path, '..'))
        for root, dirs, files in os.walk(self.path):
            rel_root = os.path.relpath(root, parent_dir).replace('\\', '/')
            for file in files:
                yield rel_root + '/' + file

    def open(self, filename, mode='r', encoding='utf-8', errors=None):
        if not os.path.isabs(filename):
            parent_dir = os.path.abspath(os.path.join(self.path, '..'))
            return self.io_open(os.path.join(parent_dir, filename), mode=mode, encoding=encoding)
        else:
            return self.io_open(filename, mode=mode, encoding=encoding)

    def close(self):
        pass


class TarExtractor(Extractor):
    def __init__(self, filename):
        self.tar = tarfile.open(filename, 'r:gz')
        self.io_open = io.open

    def names(self):
        return (info.name for info in self.tar.getmembers())

    def open(self, filename, mode='r', encoding='utf-8', errors=None):
        filename = filename.replace('\\', '/').replace('./', '')
        if not os.path.isabs(filename):
            try:
                handle = self.tar.extractfile(filename)
                return with_decoding(handle, encoding=encoding if mode != 'rb' else None)
            except KeyError:
                raise IOError('Not found in archive: {}'.format(filename))
        else:
            return self.io_open(filename, mode=mode, encoding=encoding)

    def close(self):
        self.tar.close()


class ZipExtractor(Extractor):
    def __init__(self, filename):
        self.zfile = zipfile.ZipFile(filename, 'r')
        self.io_open = io.open

    def names(self):
        return self.zfile.namelist()

    def open(self, filename, mode='r', encoding='utf-8', errors=None):
        filename = filename.replace('\\', '/').replace('./', '')
        if not os.path.isabs(filename):
            try:
                output = with_decoding(StringIO(self.zfile.read(filename).decode(encoding)), None)
                return output
            except KeyError:
                raise IOError('Not found in archive: {}'.format(filename))
        else:
            return self.io_open(filename, mode=mode, encoding=encoding)

    def close(self):
        self.zfile.close()


def extract_metadata(dist, origin=None):
    """"""
    if dist.lower().endswith('.whl'):
        result = _fetch_from_wheel(dist)
    elif dist.lower().endswith('.zip'):
        result = _fetch_from_source(dist, ZipExtractor)
    elif dist.lower().endswith('.tar.gz'):
        result = _fetch_from_source(dist, TarExtractor)
    else:
        result = _fetch_from_source(os.path.abspath(dist), NonExtractor)
    if result is not None:
        result.origin = origin
    return result


def _fetch_from_source(source_file, extractor_type):
    """

    Args:
        source_file (str): Source file
        extractor_type (type[Extractor]): Type of extractor to use

    Returns:

    """
    extractor = extractor_type(source_file)  # type: Extractor
    with closing(extractor):
        filename = os.path.basename(source_file)
        name, version = parse_source_filename(filename)
        metadata_file = None
        pkg_info_file = None
        egg_info = None
        setup_file = None

        for info_name in extractor.names():
            if info_name.lower().endswith('pkg-info') and info_name.count('/') <= 1:
                pkg_info_file = info_name
            elif info_name.endswith('.egg-info/requires.txt'):
                egg_info = info_name
            elif info_name.endswith('metadata') and info_name.count('/') <= 1:
                metadata_file = info_name

            if info_name.endswith('setup.py') and info_name.count('/') <= 1:
                setup_file = info_name
                break

        results = None
        if egg_info:
            requires_contents = ''
            try:
                requires_contents = extractor.open(egg_info, encoding='utf-8').read()
            except KeyError:
                pass
            return _parse_requires_file(requires_contents,
                                        name,
                                        version)

        if pkg_info_file:
            results = _parse_flat_metadata(extractor.open(pkg_info_file, encoding='utf-8').read())

        if (results is None or not results.reqs) and setup_file:
            if isinstance(extractor, NonExtractor):
                fake_setupdir = source_file
            else:
                fake_setupdir = tempfile.mkdtemp()
            setup_results = _parse_setup_py(name, version, fake_setupdir,
                                            extractor.relative_opener(fake_setupdir,
                                                                      os.path.dirname(setup_file)))
            if setup_results is not None:
                if version is not None:
                    setup_results.version = version
                if setup_results.name is None:
                    setup_results.name = name
                if results:
                    setup_results.version = results.version
                return setup_results
            elif results is not None:
                return results
            else:
                return None

        if metadata_file:
            return _parse_flat_metadata(extractor.open(metadata_file, encoding='utf-8').read())


def _fetch_from_wheel(wheel):
    zfile = zipfile.ZipFile(wheel, 'r')
    try:
        metadata_file = None
        infos = zfile.namelist()
        for info in infos:
            if info.lower().endswith('metadata'):
                metadata_file = info

        if metadata_file:
            return _parse_flat_metadata(zfile.read(metadata_file).decode('utf-8'))
    finally:
        zfile.close()


def _parse_flat_metadata(contents):
    name = None
    version = None
    raw_reqs = []

    for line in contents.split('\n'):
        if line.lower().startswith('name:'):
            name = line.split(':')[1].strip()
        if line.lower().startswith('version:'):
            version = utils.parse_version(line.split(':')[1].strip())
        if line.lower().startswith('requires-dist:'):
            raw_reqs.append(line.split(':')[1].strip())

    return DistInfo(name, version, list(utils.parse_requirements(raw_reqs)))


class with_decoding(object):
    def __init__(self, wrap, encoding):
        self.file = wrap
        self.encoding = encoding

    def read(self):
        results = self.file.read()
        if self.encoding:
            results = results.decode(self.encoding)
        if six.PY2:
            results = str(''.join([i if ord(i) < 128 else ' ' for i in results]))
        return results

    def readlines(self):
        results = self.file.readlines()
        if self.encoding:
            results = [result.decode(self.encoding) for result in results]
        return results

    def write(self, *args, **kwargs):
        pass

    def __iter__(self):
        if self.encoding:
            return (line.decode(self.encoding) for line in self.file)
        return iter(self.file)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def close(self):
        pass


def setup(results, *args, **kwargs):
    name = kwargs.get('name', None)
    version = kwargs.get('version', '0.0.0')
    reqs = kwargs.get('install_requires', [])
    extra_reqs = kwargs.get('extra_requires', {})

    if version is None or isinstance(version, FakeModule):
        version = '0.0.0'

    if version is not None and version != '':
        version = pkg_resources.parse_version(str(version))

    all_reqs = list(utils.parse_requirements(reqs))
    for extra, extra_req_strs in extra_reqs.items():
        cur_reqs = [utils.parse_requirement('{} ; extra=="{}"'.format(reqstr.strip(), extra))
                    for reqstr in extra_req_strs]
        all_reqs.extend(cur_reqs)

    results.append(DistInfo(name, version, all_reqs))
    return FakeModule('dist')


class FakeModule(types.ModuleType):
    call_count = 0

    def __getitem__(self, item):
        self.call_count += 1
        if self.call_count > 30:
            raise ValueError('Unintended overflow')
        return None

    def __contains__(self, item):
        return True

    def __call__(self, *args, **kwargs):
        return FakeModule(self.__name__)

    def __setitem__(self, key, value):
        pass

    def __getattr__(self, item):
        if isinstance(item, str):
            if item == '__path__':
                return ''
            elif item == '__file__':
                return os.path.join(self.__getattribute__('__name__'), '__init__.py')
            return FakeModule(item)
        else:
            return None


def fake_import_impl(name, orig_import, modname, globals=None, locals=None, fromlist=(), level=0):
    try:
        lower_modname = modname.lower()
        if 'build_ext' in modname:
            raise ImportError()

        if 'Cython' in modname:
            raise ImportError()

        if 'Cython' in modname or (
                name == lower_modname or (name + '_') in lower_modname or
                ('_' + name) in lower_modname):
            sys.modules[modname] = FakeModule(modname)
        return orig_import(modname, globals, locals, fromlist, level)
    except ImportError:
        # Skip any cython importing to improve setup.py compatibility (e.g. subprocess32)
        if 'Cython' in modname:
            raise
        if six.PY2:
            if 'asyncio' in modname:
                raise
        if (name in modname.lower() or
                '_version' in modname or
                'version' in modname or
                modname.startswith('_')):
            modparts = modname.split('.')
            for idx, mod in enumerate(modparts):
                sys.modules['.'.join(modparts[:idx + 1])] = FakeModule(mod)
            return FakeModule(modparts[-1])
        try:
            return orig_import(modname, globals, locals, fromlist, level)
        except TypeError:
            raise ImportError
        except ImportError:
            raise


@contextlib.contextmanager
def patch(module, member, new_value):
    if isinstance(module, str):
        if module not in sys.modules:
            yield
            return

        module = sys.modules[module]

    if not hasattr(module, member):
        old_member = None
    else:
        old_member = getattr(module, member)
    setattr(module, member, new_value)
    try:
        yield
    finally:
        if old_member is None:
            delattr(module, member)
        else:
            setattr(module, member, old_member)


def _remove_encoding_lines(contents):
    lines = contents.split('\n')
    lines = [line for line in lines if not (line.startswith('#') and
                                            ('-*- coding' in line or '-*- encoding' in line or 'encoding:' in line))]
    return '\n'.join(lines)


def _parse_setup_py(name, version, fake_setupdir, opener):

    old_modules = sys.modules.copy()

    # Capture warnings.warn, which is sometimes used in setup.py files
    logging.captureWarnings(True)

    results = []
    setup_with_results = functools.partial(setup, results)

    fake_import = functools.partial(fake_import_impl, name.lower(), __import__)

    with patch(sys, 'stdout', StringIO()), \
         patch(sys, 'stderr', StringIO()), \
         patch(sys, 'exit', lambda code: None), \
         patch(sys, 'getwindowsversion', lambda: (6, 0, 1)), \
         patch('__builtin__', '__import__', fake_import), \
         patch('builtins', '__import__', fake_import), \
         patch(os, 'listdir', lambda path: []), \
         patch(os, 'getcwd', lambda: '.'), \
         patch(io, 'open', opener), \
         patch(codecs, 'open', opener), \
         patch(imp, 'load_source', lambda *args, **kwargs: FakeModule('load_source')), \
         patch(setuptools, 'setup', setup_with_results), \
         patch(distutils.core, 'setup', setup_with_results):

        spy_globals = {'__file__': os.path.join(fake_setupdir, 'setup.py'),
                       '__name__': '__main__',
                       'open': opener,
                       'setup': setup_with_results}

        contents = opener('setup.py', encoding='utf-8').read()
        try:
            if six.PY2:
                contents = _remove_encoding_lines(contents)
            contents = contents.replace('print ', '')
            exec (contents, spy_globals, spy_globals)
        except Exception as ex:
             raise MetadataError(name, version, ex)
        finally:
            # Restore the old module cache
            sys.modules = old_modules

    if not results:
        return None
    return results[0]


def _parse_requires_file(contents, name, version):
    reqs = []
    sections = list(pkg_resources.split_sections(contents))
    for section in sections:
        if section[0] is None:
            reqs.extend(utils.parse_requirements(section[1]))
        elif section[0].startswith(':python_version'):
            for req in section[1]:
                reqs.append(utils.parse_requirement(req + ' ' + section[0].replace(':', ';')))

    return DistInfo(name, version, reqs)



from __future__ import print_function

import argparse
import datetime
import logging
import os
import shutil
import sys
import tempfile

from pip._vendor import pkg_resources

import qer.compile
import qer.dists
import qer.metadata
import qer.repos.pypi

from qer import utils
from qer.compile import perform_compile
from qer.config import read_pip_default_index
from qer.repos.findlinks import FindLinksRepository
from qer.repos.pypi import PyPIRepository
from qer.repos.repository import CantUseReason
from qer.repos.multi import MultiRepository
from qer.repos.solution import SolutionRepository
from qer.repos.source import SourceRepository
from qer.versions import is_possible


def _cantusereason_to_text(reason):
    if reason == CantUseReason.VERSION_NO_SATISFY:
        return 'version mismatch'
    if reason == CantUseReason.WRONG_PLATFORM:
        return 'platform mismatch {}'.format(qer.repos.repository.PLATFORM_TAG)
    if reason == CantUseReason.WRONG_PYTHON_VERSION:
        return 'python version/interpreter mismatch ({})'.format(', '.join(
            qer.repos.repository.RequiresPython.WHEEL_VERSION_TAGS))
    if reason == CantUseReason.IS_PRERELEASE:
        return 'prereleases not used'

    return 'unknown'


def _repo_as_list(repo):
    if isinstance(repo, MultiRepository):
        return repo.repositories
    return [repo]


def _generate_no_candidate_display(req, repo, dists):
    repos = _repo_as_list(repo)

    failing_node = dists[req.name]
    nodes = failing_node.reverse_deps
    constraints = failing_node.build_constraints()
    if not nodes:
        print('No package available for {}, latest is {}'.format(req.name, 'unknown'), file=sys.stderr)
    else:
        can_satisfy = is_possible(constraints)
        if not can_satisfy:
            print('No version could possibly satisfy the following requirements:', file=sys.stderr)
        else:
            print('No version of {} could satisfy the following requirements:'.format(req.name), file=sys.stderr)

        for node in nodes:
            print('   {} requires {}'.format(node, node.dependencies[failing_node]), file=sys.stderr)

        if can_satisfy:
            all_candidates = {repo: repo.get_candidates(req) for repo in repos}
            if sum(len(candidates) for candidates in all_candidates.values()) == 0:
                print('No candidates found in any of the input sources', file=sys.stderr)
            else:
                _dump_repo_candidates(req, repos)


def _dump_repo_candidates(req, repos):
    print('Found the following candidates, none of which will work:', file=sys.stderr)
    for repo in repos:
        candidates = repo.get_candidates(req)
        print('  {}:'.format(repo), file=sys.stderr)
        if candidates:
            for candidate in repo._sort_candidates(candidates):
                print('  {}: {}'.format(candidate,
                                        _cantusereason_to_text(repo.why_cant_I_use(req, candidate))),
                      file=sys.stderr)
        else:
            print('  No candidates found', file=sys.stderr)


def _create_input_reqs(input_arg, extras=None, extra_source_repos=None):
    if input_arg == '-':
        return utils.parse_requirements(sys.stdin.readlines())
    elif os.path.isdir(input_arg):
        dist = qer.metadata.extract_metadata(input_arg)
        if dist is None:
            raise ValueError('Input arg "{}" is not directory containing setup.py or requirements file'.format(input_arg))
        # source_repo = SourceRepository(input_arg)
        if extra_source_repos is not None:
            extra_source_repos.append(input_arg)
        source_dist_name = dist.name
        if extras:
            source_dist_name += '[{}]'.format(','.join(extras))
        return [utils.parse_requirement(source_dist_name)]
    else:
        return utils.reqs_from_files([input_arg])


def run_compile(input_args, extras, constraint_files, sources, find_links, index_urls,
                wheeldir, no_index, remove_source, solutions, annotate_source):
    """
    Args:
        input_args (list[str]):
        extras (Iterable[str]):
        constraint_files (list[str]):
        sources (list[str]):
        find_links (list[str]):
        index_urls (list[str]):
        wheeldir (str):
        no_index (bool):
        remove_source (bool):
        solutions (list[str]):
        annotate_source (bool):
    Returns:

    """
    if wheeldir:
        try:
            if not os.path.exists(wheeldir):
                os.mkdir(wheeldir)
        except OSError:
            pass
        delete_wheeldir = False
    else:
        wheeldir = tempfile.mkdtemp()
        delete_wheeldir = True

    extra_sources = []
    input_reqs = {
        input_arg: _create_input_reqs(input_arg, extras, extra_sources)
        for input_arg in input_args
    }
    if extra_sources:
        remove_source = True

    sources = extra_sources + sources

    constraint_reqs = {}
    if constraint_files is not None:
        constraint_reqs = {
            input_arg: _create_input_reqs(input_arg, extras, extra_sources)
            for input_arg in constraint_files
        }

    repo = build_repo(solutions, sources, find_links, index_urls, no_index, wheeldir)

    try:
        results, roots, constraints = perform_compile(input_reqs, repo, constraint_reqs=constraint_reqs)

        req_filter = None

        if remove_source:
            if not any(isinstance(r, SourceRepository) for r in _repo_as_list(repo)):
                raise ValueError('Cannot remove results from source, no source provided')

            def is_from_source(dist):
                return not hasattr(dist.metadata, 'origin') or not isinstance(dist.metadata.origin, SourceRepository)

            req_filter = is_from_source

        lines = sorted(results.generate_lines(roots, req_filter=req_filter), key=lambda x: x[0].lower())
        left_column_len = max(len(x[0]) for x in lines)
        if annotate_source:
            repo_mapping = {}
            qer_req = pkg_resources.working_set.find(utils.parse_requirement('qer'))
            print('# Compiled by Qer Requirements Compiler ({}) on {} UTC'.format(
                qer_req.version if qer_req else 'dev',
                datetime.datetime.utcnow()))
            print('#')
            print('# Inputs:')
            for input_arg in input_args:
                input_to_print = input_arg
                if os.path.exists(input_arg):
                    input_to_print = os.path.abspath(input_arg)
                print('# {}'.format(input_to_print))
            print('#')
            print('# Repositories (this annotation produced by --annotate-source):')
            for idx, repo in enumerate(_repo_as_list(repo)):
                repo_mapping[repo] = idx
                print('# [{}] {}'.format(idx, repo))
            print('')
        annotation = ''
        for line in lines:
            if annotate_source:
                req = utils.parse_requirement(line[0])
                key = req.name + ('[{}]'.format(req.extras[0]) if req.extras else '')
                source = results[key].repo
                if not source in repo_mapping:
                    print('No repo for {}'.format(line), file=sys.stderr)
                annotation = '[{}] '.format(repo_mapping[source])
            print('{}  # {}{}'.format(line[0].ljust(left_column_len), annotation, line[1]))
    except qer.repos.repository.NoCandidateException as ex:
        _generate_no_candidate_display(ex.req, repo, ex.results)
        sys.exit(1)
    finally:
        if delete_wheeldir:
            shutil.rmtree(wheeldir)


def build_repo(solutions, sources, find_links, index_urls, no_index, wheeldir):
    repos = []
    if solutions:
        repos.extend(SolutionRepository(solution) for solution in solutions)
    if sources:
        repos.extend(SourceRepository(source) for source in sources)
    if find_links:
        repos.extend(FindLinksRepository(find_link) for find_link in find_links)
    if not no_index:
        if not index_urls:
            default_index_url = read_pip_default_index() or 'https://pypi.org/simple'
            repos.append(PyPIRepository(default_index_url, wheeldir))
        else:
            repos.extend(PyPIRepository(index_url, wheeldir) for index_url in index_urls)
    if not repos:
        raise ValueError('At least one Python distributions source must be provided.')
    if len(repos) > 1:
        repo = MultiRepository(*repos)
    else:
        repo = repos[0]
    return repo


class IndentFilter(logging.Filter):
    def filter(self, record):
        depth = getattr(record, 'depth', 0)
        record.msg = (' ' * depth) + record.msg
        return record


def compile_main():
    logging.basicConfig(level=logging.ERROR)

    parser = argparse.ArgumentParser()
    parser.add_argument('requirement_files', nargs='+', help='Input requirements files')
    parser.add_argument('-v', '--verbose', default=False, action='store_true',
                        help='Enable verbose output to stderr')
    parser.add_argument('-c', '--constraints', action='append',
                        help='Contraints files. Not included in final compilation')
    parser.add_argument('-e', '--extra', nargs='+', dest='extras', default=[],
                        help='Extras to apply automatically to source packages')
    parser.add_argument('--remove-source', default=False, action='store_true',
                        help='Remove distributions satisfied via --source from the output')
    parser.add_argument('--annotate-source', default=False, action='store_true',
                        help='Annotate the output file with the sources of each requirement')

    add_repo_args(parser)

    args = parser.parse_args()
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG, stream=sys.stderr)
        logging.getLogger('qer').setLevel(logging.DEBUG)

        logging.getLogger('qer.compile').addFilter(IndentFilter())

    run_compile(args.requirement_files, args.extras, args.constraints if args.constraints else None, args.sources, args.find_links,
                args.index_urls, args.wheel_dir, args.no_index, args.remove_source, args.solutions, args.annotate_source)


def add_repo_args(parser):
    parser.add_argument('-i', '--index-url', nargs='+', dest='index_urls', default=[])
    parser.add_argument('-f', '--find-links', nargs='+', default=[])
    parser.add_argument('-s', '--source', nargs='+', dest='sources', default=[])
    parser.add_argument('-w', '--wheel-dir', type=str, default=None)
    parser.add_argument('-u', '--solution', nargs='+', dest='solutions', default=[],
                        help='Existing fully-pinned constraints file to use as a baseline when compiling')
    parser.add_argument('--no-index', action='store_true', default=False,
                        help='Do not connect to the internet to compile')


if __name__ == '__main__':
    compile_main()
